use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use prost::Message;

///
pub struct Service {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
}

impl std::fmt::Debug for Service {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("beteran-frontend-server-edge::identity::service::Service").finish()
  }
}

impl Service {
  ///
  pub fn new(connection_broker: nats::asynk::Connection, queue_broker: String) -> Service {
    Service {
      connection_broker,
      queue_broker,
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(
      self.check_username_for_duplication(),
      self.check_nickname_for_duplication(),
      self.captcha(),
      self.signin(),
    )
    .map(|_| ())
  }

  fn get_client_in_header(
    &self,
    message: &nats::asynk::Message,
  ) -> Result<bpr::models::core::network::Client, bcr::error::rpc::Error> {
    match &message.headers {
      Some(headers) => {
        let client = match headers.get(bpr::c2se::core::network::HEADER_CLIENT) {
          Some(c) => {
            let msg = base64::decode(c).map_err(|e| {
              bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
                message: format!("invalid header: {}", e),
              })
            })?;
            bpr::models::core::network::Client::decode(msg.as_slice()).map_err(|e| {
              bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
                message: format!("invalid header: {}", e),
              })
            })?
          }
          None => {
            return Err(bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
              message: "invalid client information".to_string(),
            }));
          }
        };

        Ok(client)
      }
      None => Err(bcr::error::rpc::Error::Parse(bcr::error::rpc::Parse {
        message: "invalid header".to_string(),
      })),
    }
  }

  async fn check_username_for_duplication(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::identity::SUBJECT_CHECK_USERNAME_FOR_DUPLICATION,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;

        let req = bpr::c2se::common::identity::CheckUsernameForDuplicationRequest::decode(
          message.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
            message: format!("invalid request: {}", e),
          })
        })?;

        let ss_check_username_for_duplication_req =
          bpr::ss::member::identity::CheckUsernameForDuplicationRequest {
            client: Some(client),
            username: req.username,
          };

        let ss_check_username_for_duplication_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member::identity::SUBJECT_CHECK_USERNAME_FOR_DUPLICATION,
            ss_check_username_for_duplication_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        let ss_check_username_for_duplication_res =
          bpr::ss::member::identity::CheckUsernameForDuplicationResponse::decode(
            ss_check_username_for_duplication_res_msg.data.as_slice(),
          )
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        if let Some(e) = ss_check_username_for_duplication_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server: {}", e),
          }));
        }

        if let Some(r) = ss_check_username_for_duplication_res.result {
          message
            .respond(
              bpr::c2se::common::identity::CheckUsernameForDuplicationResponse {
                error: None,
                result: Some(
                  bpr::c2se::common::identity::check_username_for_duplication_response::Result {
                    duplicated: r.duplicated,
                  },
                ),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                message: format!("server: {}", e),
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::common::identity::CheckUsernameForDuplicationResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn check_nickname_for_duplication(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::identity::SUBJECT_CHECK_NICKNAME_FOR_DUPLICATION,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;

        let req = bpr::c2se::common::identity::CheckNicknameForDuplicationRequest::decode(
          message.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
            message: format!("invalid request: {}", e),
          })
        })?;

        let ss_check_nickname_for_duplication_req =
          bpr::ss::member::identity::CheckNicknameForDuplicationRequest {
            client: Some(client),
            nickname: req.nickname,
          };

        let ss_check_nickname_for_duplication_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member::identity::SUBJECT_CHECK_NICKNAME_FOR_DUPLICATION,
            ss_check_nickname_for_duplication_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        let ss_check_nickname_for_duplication_res =
          bpr::ss::member::identity::CheckNicknameForDuplicationResponse::decode(
            ss_check_nickname_for_duplication_res_msg.data.as_slice(),
          )
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        if let Some(e) = ss_check_nickname_for_duplication_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server: {}", e),
          }));
        }

        if let Some(r) = ss_check_nickname_for_duplication_res.result {
          message
            .respond(
              bpr::c2se::common::identity::CheckNicknameForDuplicationResponse {
                error: None,
                result: Some(
                  bpr::c2se::common::identity::check_nickname_for_duplication_response::Result {
                    duplicated: r.duplicated,
                  },
                ),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                message: format!("server: {}", e),
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::common::identity::CheckNicknameForDuplicationResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn captcha(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::identity::SUBJECT_CAPTCHA,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;

        let _req = bpr::c2se::common::identity::CaptchaRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_captcha_req = bpr::ss::member::identity::CaptchaRequest {
          client: Some(client),
        };

        let ss_captcha_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member::identity::SUBJECT_CAPTCHA,
            ss_captcha_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        let ss_captcha_res =
          bpr::ss::member::identity::CaptchaResponse::decode(ss_captcha_res_msg.data.as_slice())
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                message: format!("server: {}", e),
              })
            })?;

        if let Some(e) = ss_captcha_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server: {}", e),
          }));
        }

        if let Some(r) = ss_captcha_res.result {
          message
            .respond(
              bpr::c2se::common::identity::CaptchaResponse {
                error: None,
                result: Some(bpr::c2se::common::identity::captcha_response::Result {
                  token: r.token,
                  image: r.image,
                }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                message: format!("server: {}", e),
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::common::identity::CaptchaResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn signin(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::c2se::frontend::identity::SUBJECT_SIGNIN,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let client = self.get_client_in_header(&message)?;

        let req = bpr::c2se::frontend::identity::SigninRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let ss_signin_req = bpr::ss::member::identity::SigninRequest {
          client: Some(client),
          token: req.token,
          security_code: req.security_code,
          username: req.username,
          password: req.password,
        };

        let ss_signin_res_msg = self
          .connection_broker
          .request(
            bpr::ss::member::identity::SUBJECT_SIGNIN,
            ss_signin_req.encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        let ss_signin_res =
          bpr::ss::member::identity::SigninResponse::decode(ss_signin_res_msg.data.as_slice())
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                message: format!("server: {}", e),
              })
            })?;

        if let Some(e) = ss_signin_res.error {
          return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server: {}", e),
          }));
        }

        if let Some(r) = ss_signin_res.result {
          message
            .respond(
              bpr::c2se::frontend::identity::SigninResponse {
                error: None,
                result: Some(bpr::c2se::frontend::identity::signin_response::Result {
                  session_id: r.session_id,
                }),
              }
              .encode_to_vec(),
            )
            .await
            .map_err(|e| {
              bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
                message: format!("server: {}", e),
              })
            })?;
        }

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::c2se::frontend::identity::SigninResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }
}
